
#include "AD7606Control.h"
#include "timer_private.h"

// Schnitstelle zur Ausgabe
SerialControl* outDev;

// Schnittstelle zum AD7606
UrsAD7606Spi<rstPin, cvaPin, busyPin, csPin, AD7606Range::Volt5> ursAD7606Spi;

// Timer, der die Messungen innerhalb einer Messreihe taktet
// Wenn ein anderer Timer verwendet wird, muss ein anderer Interrupts berschrieben werden.
HardwareTimer sampleTimer = HardwareTimer(1);

// Messreihe definieren
int sampleCount = 10;                // Anzahl Messungen, die zu einer Messreihe gehren (< maxSampleCount)
AD7606DataSet samples[maxSampleCount]; // Puffer zur Aufnahme der Messwerte
volatile bool samplingDone = false;  // true, wenn eine Messreihe beendet wurde
volatile int currentSampleCount = 0; // Aktuell gestartete Messungen

volatile int currentDestCount = 0; // Aktuell eingelesene Messungen

// Interrupthandler
// Die ISRs sind in der Lib in einer .c-Datei als __weak deklariert. berschreiben nur mit C-Namens-Konvention.
extern "C" {
   // ISR fr Timer1-Upload-Interrupt
   void __irq_tim1_up(void) {
      sampleTimer.c_dev()->regs.adv->SR = ~TIMER_SR_UIF; // Timer-Interrupt-Flag lschen.

      if (currentSampleCount++ == sampleCount) { // Alle Messungen wurden gestartet.
         sampleTimer.c_dev()->regs.adv->CR1 &= ~TIMER_CR1_CEN;  //Timer abschalten.
         return;
      }

      ursAD7606Spi.start(); // neue Messung starten.
   }


   // ISR fr PA1-Interrupt (BUSY). Gleichzeitig PB1, PC1, ... Diese sind aber nicht aktiviert.
   void __irq_exti1(void) {
      EXTI_BASE->PR = (1U << EXTI1); // Interrupt-Flag lschen.
      ursAD7606Spi.startReading();   // Auslesen starten.
   }


   // ISR fr DMA1-Channel2-Transfer-Complete-Interrupt. Gleichzeitig auch Transfer-Error-Interrupt, ... Diese sind aber nicht aktiviert.
   void __irq_dma1_channel2(void) {
      ursAD7606Spi.endTransfer();

      if (currentDestCount == sampleCount-1) { // Alle Messungen wurden eingelesen.
         samplingDone = true; // Messreihe komplett.
         return;
      }

      // Weitere Messungen stehen aus.
      ursAD7606Spi.beginTransfer(&samples[++currentDestCount]);
   }

} // extern "C"

// Das freigeben der Interrupts ber attach.. erfordert den Verweis auf einen Handler != NULL.
void DummyHandler() {}


void AD7606Control::begin(SerialControl* sc) {
   outDev = sc;

   ursAD7606Spi.begin(AD7606DmaTci::enable, AD7606DummyRead::enable);

   sampleTimer.init();
   sampleTimer.setPrescaleFactor(1); // Timer-Takt ist 72MHz
   setTime(8); // Taktzeit einstellen, Default ist 8us

   setCount(10); // Gre der Messreihe, Default ist 10 Samples pro Messreihe

   sampleTimer.attachInterrupt(0, DummyHandler); // Soll nur den Interrupt freigeben
   attachInterrupt(busyPin, DummyHandler, FALLING); // Soll nur den Interrupt freigeben
   dma_attach_interrupt(DMA1, DMA_CH2, DummyHandler); // Soll nur den Interrupt freigeben

   // Interrupts mit hoher Prioritt versehen
   nvic_irq_set_priority(NVIC_TIMER1_UP_TIMER10, 0);
   nvic_irq_set_priority(NVIC_EXTI1, 0);
   nvic_irq_set_priority(NVIC_DMA_CH3, 0);
}

void AD7606Control::setTime(uint microSeconds) {
   if (microSeconds < minTime)
      microSeconds = minTime;
   sampleTimer.setOverflow(microSeconds * 72); // Timer-Takt ist 76MHz

   outDev->print("Timer set to "); outDev->print(microSeconds); outDev->println(" us.");
}

void AD7606Control::setCount(uint sc) {
   if (sc > maxSampleCount)
      sc = maxSampleCount;
   sampleCount = sc;
   outDev->print("Sample Count set to "); outDev->print(sampleCount); outDev->println(".");
}

void AD7606Control::startSampling() {
   outDev->println("Start sampling");
   Serial.flush(); // Seielle bertragung abschlieen.
   currentDestCount = 0; // Neue Messreihe.

   sampleTimer.c_dev()->regs.adv->CNT = 58; // Zeit fr start.
   sampleTimer.c_dev()->regs.gen->CR1 |= TIMER_CR1_CEN;  //Timer anschalten.

   ursAD7606Spi.start();
   ursAD7606Spi.beginTransfer(&samples[currentDestCount]); // Ziel fr erste Messung.
   currentSampleCount = 1; // Erste Messung gestartet.
}

void AD7606Control::reset() {
   ursAD7606Spi.reset(AD7606DummyRead::enable);
   outDev->println("Reset done");
}

bool AD7606Control::hasFinished() {
   bool temp = samplingDone;
   samplingDone = false;
   return temp;
}

void AD7606Control::printReport() {
   outDev->println("     Ch1      Ch2      Ch3      Ch4      Ch5      Ch6      Ch7      Ch8");
   for (int s = 0; s < sampleCount; s++) { // ber alle Messreihen

      // Messreihennummer ausgeben.
      int n = s + 1;
      if (n < 10)outDev->print(" ");
      if (n < 100)outDev->print(" ");
      outDev->print(n); outDev->print(" ");

      // Messungen ausgeben.
      for (int i = 0; i < 8; i++) { // ber alle Kanle
         auto d = ursAD7606Spi.value(samples[s].channel[i]);
         if (d >= 0)
            outDev->print(" "); // Platz fr fehlendes Minus-Zeichen
         outDev->print(d, 4);
         outDev->print("  ");
      }
      outDev->println("");
   }
}

