﻿/*
 Name:		UrsAD7606Spi
 Created:	24.11.2021
 Author:	   Ulli

 Version 1.0 (2021-11-26)
 -------------------------
 - Basis-Version

*/

#ifndef _URSAD7606SPI_h
#define _URSAD7606SPI_h

#include <Arduino.h>
#include <UrsSTM32F1-AD7606.h>
#include <libmaple/gpio.h>
#include <libmaple/spi.h>
#include <libmaple/dma.h>
#include <board/board.h>

extern spi_dev* const ad7606SpiDev;                // = SPI1; Einlesen erfolgt über SPI1. Bei SPI2 ist ein anderer ClockDivider notwendig!
extern dma_dev* const ad7606DmaDev;                // = DMA1; SPI1 wird von DMA1 bedient
extern gpio_dev* const ad7606pinRdDev;             // = GPIOA; Beim Bluepill liegt SPI1-SCLK auf PA5
extern dma_tube_reg_map* const ad7606TxDmaChannel; // = DMA1CH3_BASE; SPI1-TX-DMA-Kanal
extern dma_tube_reg_map* const ad7606RxDmaChannel; // = DMA1CH2_BASE; SPI1-RX-DMA-Kanal
extern uint16_t transmitBuf[1];                    // = {0}; nur formal. Es wird nichts gesendet
const dma_channel ad7606TxDmaChannelNo = DMA_CH3;  // Kanalnummer für Kanal 3 (TX)
const dma_channel ad7606RxDmaChannelNo = DMA_CH2;  // Kanalnummer für Kanal 2 (RX)
const uint32_t ad7606DmaTICFMask = DMA_ISR_TCIF3;  // Maske für das TCIF-Flag auf Kanal 3 (TX)
const uint32_t ad7606ChannelsToRead = 8;           // Anzahl Kanäle, die eingelesen werden sollen.

const uint8_t pinRD = BOARD_SPI1_SCK_PIN;    // Pins für SPI1: RD beim AD7606, SCLK bei SPI
const uint8_t pinMISO = BOARD_SPI1_MISO_PIN; // Pins für SPI1: DoutA beim AD7606, MISO bei SPI

extern volatile uint32_t* pinCsBSRR;  // Pointer auf das BSSR-Register des GPIO-Device für den CS-Pin
extern volatile uint32_t* pinCvaBSRR; // Pointer auf das BSSR-Register des GPIO-Device für den CVA-Pin
extern volatile uint32_t* pinBusyIdr; // Pointer auf das IDR-Register des GPIO-Device für den BUSY-Pin


// Steuerung des AD706 mit byte-serieller Datenübertragung
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
class UrsAD7606Spi {
public:
   // Initilialisiert die Pins und das SPI-Interface
   static void begin(AD7606DmaTci enableRxTCI = AD7606DmaTci::disable, AD7606DummyRead dummyRead = AD7606DummyRead::enable);

   // Gibt das SPI- und DMA-Interface frei
   static void end(bool freePins = true); // true: die GPIOs werden auf INPUT zurück gestellt.

   // Führt einen Reset des AD7606 durch
   static void reset(AD7606DummyRead dummyRead = AD7606DummyRead::enable);

   // Startet eine Messung.
   static inline void start();

   // Liest die Messdaten synchron (blockierend) aus.
   static inline void readData(AD7606DataSet* dest);

   // Startet eine Messung und wartet, bis die Messung abgeschlossen ist.
   // Steuerung über BUSY.
   static inline void sample();

   // Startet eine Messung, wartet, bis die Messung abgeschlossen ist, und liest anschließend die Daten ein.
   static inline void sampleAndRead(AD7606DataSet* dest);

   // Liefert true, solange ein Messvorgang läuft.
   static inline bool isBusy();

   // Rohwert in Spannung umrechnen.
   static float value(int16_t rawValue) {
      return (float)rawValue * conversion;
   }

   // Elemetare Übertagungsfunktionen
   static inline void beginTransfer(AD7606DataSet* dest); // Bereitet SPI und DMA für den nächsten Zyklus vor.
   static inline void startReading(); // Startet die SPI-DMA-Datenübertagung.
   static inline void endTransfer();  // Übertragung abschließen, Register zurückstellen.


protected:
   // float conversion; // 1 LSB = 152.58 μV bei 5V, 1 LSB = 305.175 μV bei 10V
   constexpr static float conversion = (float)range / (float)1000000000;

   static void initSpi();  // Initialisiert die SPI-Register
   static void initDma(AD7606DmaTci enableRxTCI);  // Initialisiert die DMA-Register
};

/*=======================================================
  === Implementierung                                 ===
  =======================================================*/

template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::begin(AD7606DmaTci enableRxTCI, AD7606DummyRead dummyRead) {
   // init GPIO
   pinMode(pinRST, OUTPUT);
   digitalWrite(pinRST, LOW); // Ruhezustand für RESET ist LOW.

   pinMode(pinCVA, OUTPUT);
   digitalWrite(pinCVA, LOW); // Ruhezustand für CVA/CVB ist LOW.

   pinMode(pinBUSY, INPUT);   // BUSY ist ein Eingangsignal.

   gpio_set_mode(ad7606pinRdDev, pinRD, GPIO_AF_OUTPUT_PP); // Keine passende pinMode-Variante
   pinMode(pinMISO, INPUT);

   pinMode(pinCS, OUTPUT);
   digitalWrite(pinCS, HIGH); // Ruhezustand für CS ist HIGH.


   pinCsBSRR = &(PIN_MAP[pinCS].gpio_device->regs->BSRR);
   pinCvaBSRR = &(PIN_MAP[pinCVA].gpio_device->regs->BSRR);
   pinBusyIdr = &(PIN_MAP[pinBUSY].gpio_device->regs->IDR);
   initSpi();
   initDma(enableRxTCI);

   reset(dummyRead); // Reset des AD7606
}

// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::initSpi() {
   spi_init(ad7606SpiDev); // Initialisiert RCC

// siehe RM0008 Reference Manual Abschnitt 25.5.1 SPI control register 1 (SPI_CR1)
   ad7606SpiDev->regs->CR1
      = SPI_MODE_2              // CPHA, CPOL
      | SPI_CR1_MSTR            // MSTR: Master configuration
      | SPI_CR1_BR_PCLK_DIV_2   // 36 MHz bei 72 MHz F_CPU, bei SPI2 nur 18 MHz
      | SPI_CR1_SPE             // SPE:      1 = SPI enabled 
      // | SPI_CR1_LSBFIRST     // LSBFIRST: 0 = MSB transmitted first
      | SPI_CR1_SSI             // SSI (Internal slave select)
      | SPI_CR1_SSM             // SSM (Software slave management)
      // | SPI_CR1_RXONLY       // RXONLY:   0 = Output ensabled
      | SPI_CR1_DFF_16_BIT      // DFF:      1 = 16-bit data frame format 
      // | SPI_CR1_CRCNEXT      // CRCNEXT:  0 = no CRC phase
      // | SPI_CR1_CRCEN        // CRCEN:    0 = CRC calculation disabled
      // | SPI_CR1_BIDIOE       // BIDIOE:   0 = Output disabled (receive-only mode), keine Wirkung
      | SPI_CR1_BIDIMODE_2_LINE // BIDIMODE: 0 = 2-line unidirectional data mode selected
      ;

   // siehe RM0008 Reference Manual   Abschnitt 25.5.2 SPI control register 1 (SPI_CR2)
   ad7606SpiDev->regs->CR2
      = 0
      // | SPI_CR2_RXDMAEN   // RXDMAEN: 0 = Rx buffer DMA disabled
      // | SPI_CR2_TXDMAEN   // TXDMAEN: 0 = Tx buffer DMA disabled
      // | SPI_CR2_SSOE      // SSOE:    0 = SS output is disabled in master mode (1 klappt nicht!)
      // | SPI_CR2_ERRIE     // ERRIE:   0 = Error interrupt is masked (disabled)
      // | SPI_CR2_RXNEIE    // RXNEIE:  0 = RXNE interrupt masked (disabled)
      // | SPI_CR2_TXEIE     // TXEIE:   0 = TXE interrupt masked (disabled)
      ;
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::initDma(AD7606DmaTci enableRxTCI) {
   dma_init(ad7606DmaDev); // Initialisiert RCC

   uint32_t DMA_CIE = enableRxTCI == AD7606DmaTci::enable ? DMA_CCR_TCIE : 0; // Bit für TCIE

   // siehe RM0008 Reference Manual   Abschnitt 13.4.3 DMA channel x configuration register (DMA_CCRx)
   ad7606RxDmaChannel->CCR
      = 0
      //  | DMA_CCR_EN           // disable DMA, can't write to CMAR/CPAR otherwise 
      //  | DMA_CCR_TCIE         // TCIE: 0 = TC interrupt disabled, wird nachträglich eingestellt
      //  | DMA_CCR_HTIE         // HTIE: 0 = HT interrupt disabled
      //  | DMA_CCR_TEIE         // TTIE: 0 = TE interrupt disabled
      //  | DMA_CCR_DIR_FROM_PER // DIR: 0 = Read from peripheral
      //  | DMA_CCR_CIRC         // CIRC: 0 = Circular mode disabled
      //  | DMA_CCR_PINC         // PINC: 0 = Peripheral increment mode disabled
      | DMA_CCR_MINC             // MINC: 1 = Memory increment mode enabled
      | DMA_CCR_PSIZE_16BITS     // PSIZE: Peripheral size 16 Bit
      | DMA_CCR_MSIZE_16BITS     // MSIZE: Memory size 16 Bit
      | DMA_CCR_PL_VERY_HIGH     // PL: Channel priority level very high
      //  | DMA_CCR_MEM2MEM      // MEM2MEM: 0 = Memory to memory mode disabled
      ;
   ad7606RxDmaChannel->CCR |= DMA_CIE; // Ggf. Transfer-Complete-Interrrupt freigeben
   ad7606RxDmaChannel->CPAR = (uint32)&ad7606SpiDev->regs->DR;
   ad7606RxDmaChannel->CNDTR = 8; // 8 Dateneinheiten zu 16 Bit müssen übertragen werden

   ad7606TxDmaChannel->CCR
      = 0
      // | DMA_CCR_EN          // disable DMA, can't write to CMAR/CPAR otherwise 
      // | DMA_CCR_TCIE           // TCIE: 1 = TC interrupt enabled
      // | DMA_CCR_HTIE           // HTIE: 1 = HT interrupt enabled
      // | DMA_CCR_TEIE        // TTIE: 0 = TE interrupt disabled
      | DMA_CCR_DIR_FROM_MEM   // DIR: 1 = Read from Memory
      // | DMA_CCR_CIRC        // CIRC: 0 = Circular mode disabled
      // | DMA_CCR_PINC        // PINC: 0 = Peripheral increment mode disabled
      // | DMA_CCR_MINC        // MINC: 0 = Memory increment mode disabled
      | DMA_CCR_PSIZE_16BITS   // PSIZE: Peripheral size 16 Bit
      | DMA_CCR_MSIZE_16BITS   // MSIZE: Memory size 16 Bit
      | DMA_CCR_PL_VERY_HIGH         // PL: Channel priority level low
      // | DMA_CCR_MEM2MEM     // MEM2MEM: 0 = Memory to memory mode disabled
      ;
   ad7606TxDmaChannel->CPAR = (uint32)&ad7606SpiDev->regs->DR;
   ad7606TxDmaChannel->CNDTR = 8; // 8 Dateneinheiten zu 16 Bit müssen übertragen werden
   ad7606TxDmaChannel->CMAR = (uint32)transmitBuf; // immer 0x0 senden
}

// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::end(bool freePins) {
   ad7606SpiDev->regs->CR1 = 0; // Reset-Wert
   ad7606SpiDev->regs->CR2 = 0; // Reset-Wert

   ad7606TxDmaChannel->CCR = 0; // Reset-Wert
   ad7606RxDmaChannel->CCR = 0; // Reset-Wert
   // // dma_clear_isr_bits
   ad7606DmaDev->regs->IFCR = (1U << (4 * (ad7606TxDmaChannelNo - 1))) | (1U << (4 * (ad7606RxDmaChannelNo - 1)));

   if (freePins) {
      pinMode(pinRST, INPUT);
      digitalWrite(pinRST, LOW);

      pinMode(pinCVA, INPUT);
      digitalWrite(pinCVA, LOW);

      pinMode(pinRD, INPUT);
      digitalWrite(pinRD, LOW);

      pinMode(pinMISO, INPUT);
      digitalWrite(pinMISO, LOW);

      pinMode(pinCS, INPUT);
      digitalWrite(pinCS, LOW);
   }
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::readData(AD7606DataSet* dest) {
   beginTransfer(dest); // SPI und DMA vorbereiten

   startReading(); // Übertragung starten.

   // Warten bis die Übertragung abgeschlossen ist. Das TICF-Flag auf dem TX-DMA-Kanal gesetzt wird.
   while ((ad7606DmaDev->regs->ISR & ad7606DmaTICFMask) == 0) {}

   // Warten bis TXE=1 und dann bis BSY=0 bevor die SPI deaktiviert wird.
   while (spi_is_tx_empty(ad7606SpiDev) == 0); // Warten bis TXE=1 (spi_is_tx_empty() ist als inline deklariert).
   while (spi_is_busy(ad7606SpiDev) != 0);     // Warten bis BSY=0 (spi_is_busy() ist als inline deklariert).

   endTransfer(); // SPI und DMA abschalten.
}



// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::beginTransfer(AD7606DataSet* dest) {
   *pinCsBSRR = (1U << (pinCS)) << 16;; // CS auf LOW.

   // DMA konfigurieren
   ad7606RxDmaChannel->CMAR = (uint32_t)dest; // Ziel-Adresse
   ad7606RxDmaChannel->CNDTR = ad7606ChannelsToRead; // Anzahl Kanäle, die eingelesen werden sollen.
   ad7606TxDmaChannel->CNDTR = ad7606ChannelsToRead;
   ad7606RxDmaChannel->CCR |= DMA_CCR_EN; // DMA anschalten.
   ad7606TxDmaChannel->CCR |= DMA_CCR_EN;
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::endTransfer() {
   // DMA muss abgeschaltet werden, ansonsten können die notwendigen Register nicht beschrieben werden.

   ad7606SpiDev->regs->CR2 &= ~(SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN); // DMA im SPI-Device abschalten.

   ad7606RxDmaChannel->CCR &= ~DMA_CCR_EN; // DMA abschalten.
   ad7606TxDmaChannel->CCR &= ~DMA_CCR_EN;

   dma_clear_isr_bits(ad7606DmaDev, ad7606TxDmaChannelNo); // Interrupt-Flags löschen.
   dma_clear_isr_bits(ad7606DmaDev, ad7606RxDmaChannelNo); // Interrupt-Flags löschen.

   *pinCsBSRR = (1U << (pinCS)); // CS auf HIGH.
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::startReading() {
   // DMA im SPI-Device anschalten
   ad7606SpiDev->regs->CR2 |= SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN; // Erzeugt ersten DMA-Request.
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::sampleAndRead(AD7606DataSet* dest) {
   start();
   while (isBusy()) {}
   readData(dest);
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::start() {
   *pinCvaBSRR = (1U << (pinCVA));       // Bit im Low-Word des BSRR-Registers setzen [digitalWrite(pinCVA, HIGH);].
   *pinCvaBSRR = (1U << (pinCVA)) << 16; // Bit im High-Word des BSRR-Registers setzen [digitalWrite(pinCVA, LOW);].


}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::sample() {
   start();
   while (isBusy());
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline bool UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::isBusy() {
   return ((*pinBusyIdr & (1 << pinBUSY)) != 0);
}


// =========================================================================================================================
template<uint8_t pinRST, uint8_t pinCVA, uint8_t pinBUSY, uint8_t pinCS, AD7606Range range>
inline void UrsAD7606Spi<pinRST, pinCVA, pinBUSY, pinCS, range>::reset(AD7606DummyRead dummyRead) {
   digitalWrite(pinRST, HIGH);
   delayMicroseconds(1); // minimal 50 ns (tRESET im Datenblatt)
   digitalWrite(pinRST, LOW);

   if (dummyRead == AD7606DummyRead::enable) {// einen nicht ausgewerteten Read vornweg.
      sample(); // Messung starten und warten bis sie abgeschlossen ist.
   }
}
#endif